<?php

//
// Tables
//
$g_ITP = array();
$g_MOBNAME = array();
$g_SKILL = array();
$g_CONST = array();
$g_OP = array();


// returns: array (const -> id)
function read_def($filename){

	print "Loading $filename... ";

	$content = @file_get_contents($filename);
	if($content === false)
		die("read fail (IO Error)\n");
	
	
	// split
	$arr = explode("\n", $content);
	$szArr = sizeof($arr);
	$ret = array();

	$cnt = 0;	
	for($i = 0;  $i < $szArr;  $i++){
		$tmp = explode(' ', $arr[$i]);
		if(sizeof($tmp) != 2)
			continue;
		
		$ret[ $tmp[0] ] = (int)str_replace("\r", '', $tmp[1] );
		$cnt++;		
		
		unset($tmp);
	}
	
	print "done ($cnt definitions)\n";
	

	unset($content, $arr);

return $ret;
}//end read_def()



function read_enum_v2(){
	global $g_CONST;
	print "Loading scriptdata/enum_v2.sc... ";
	
	$fp = @fopen('scriptdata/enum_v2.sc', 'r');
	if(!$fp)
		die("read fail (IO open Error)\n");
	//
	$nConst = 0;	
	
	// read first line 
	// as its the version date 
	//
	$version = fgets($fp, 1024);
	if($version === false || $version === null)
		die("cannot get version date (first line)\n");
	
	$version = rtrim($version);
	if(strcmp($version, '20030414') !== 0)
		die("invalid version, expected 20030414\n");
	

	// 	
	$n = 0; // counter
	$linecnt = 0;
	while(1){
		$linecnt++;
		$line = fgets($fp, 8192);
		if($line === false || $line === null)
			break;
		
		$line = trim($line);
		
		if(strlen($line) == 0)
			continue;
		
		$c = $line[0];	
		
		if( $c === ';' || ($c == '/' && $line[1] == '/') ){
			continue; // comment
				
		}else if($c === '#'){
			//
			// Reset counter Instruction (orig parser strips all chars until it finds alnum) -> trim() will do its job here.
			// 
			$n = (int)trim(substr($line, 1));
		}else{
			// tokens are sepperated using ',' 
			
			$arr = explode(',', $line);		// when token for splitting isnt found, php will return $line as $arr[0], as wanted by enum.sc
			foreach($arr as $x => &$token){
				$g_CONST[ trim($token) ] = $n++;

				$nConst++;
			}
			
			
		}
		
	
	}//endwhile


	print "done ($nConst constants)\n";
	fclose($fp);
}//end: read_enum_v2()



function read_common_asm(){
	global $g_CONST, $g_OP;
	
	print "Loading scriptbin/common.asm... ";
	
	$fp = @fopen('scriptbin/common.asm', 'r');
	if(!$fp)
		die("read fail (IO open Error)\n");
	
	//
	$last = null;
	$nOp = 0; $nConst = 0;		

	$linecnt = 0;
	while(1){
		$linecnt++;
		$line = fgets($fp, 1024);
		if($line === false || $line === null)
			break;
		
		$linelen = strlen($line);
		if($linelen === 0)
			continue;	
		
		if($line[0] != '/' || $line[1] != '/')
			continue; // skip non comment lines
		
		// remove linebreaks 
		$line = rtrim($line);
		
		// 
		$arr = explode(' ', $line);
		array_shift($arr);	// remove comment
		
		$cmd = array_shift($arr);
		if(strcmp($cmd, 'define') === 0){
			//
			// define
			//
			if(sizeof($arr) != 2) // syntax error
				die("line $linecnt, ddefine expects exact 2 parameters <string> <int>\n");
				
			// define const id
			$g_CONST[ $arr[0] ] = (int)$arr[1];	
			$nConst++;
		
			// as last defintion wasnt a declare!
			$last = null;
		
		}else if(strcmp($cmd, 'declare') === 0){
			//
			// declare
			//
			if(sizeof($arr) != 3)	// syntax error
				die("line $linecnt, declare expects eact 3 parameters <string> <string> <int>\n");
			
			$token = array_shift($arr);
			$param = array_shift($arr);
			$op = array_shift($arr);
			
			// build params array (in order for parsing)
			$tmp = array();
			$va = false;
			$va_type = null;
			for($i = 0; $i < strlen($param); $i++){
				$type = $param[$i];
				
				if($type === 's' || $type === 'n')
					$tmp[] = $type;
				else if($type === '.'){
					if($i+1 !== strlen($param))
						die("line $linecnt, declare of $token got . (no arguments) with following type specifiers?\n");
					break;
				}else if($type === '?'){
					if($i === 0)
						die("line $linecnt, declare of $token with VA Arg, but with no leading parameter?!\n");
					
					if($i+1 !== strlen($param))
						die("line $linecnt, declare fo $token with VA Arg, after VA no more arguments can be specified!\n");
					
					$va = true;
					$va_type = $param[$i-1];
				}
								
			}
			
			// 
			// build op description
			//
			$def = array('name' => $token,
						 'blockcheck' => false,
						 'param' => $tmp,
						 'nParam' => sizeof($tmp),
						 'va' => $va,
						 'va_type' => $va_type);
			
			// register OP
			$g_OP[ $token ] = $def;
			
			$last = &$g_OP[ $token ];
				
			//
			$nOp++;
		
		}else if(strcmp($cmd, 'block') === 0){
			//
			// block check func
			//
			if(sizeof($arr) != 2)
				die("line $linecnt, block expects 2 parameters <string> <string>\n");
			
			if(strcmp($arr[0], 'check') !== 0 || strcmp($arr[1], 'func') !== 0)
				die("line $linecnt, block -> only check func supported!\n");
			
			// set last func to block check func!
			if($last === null)
				die("line $linecnt, block check func set, but no prepending declaration found!\n");
			
			//
			$last['blockcheck'] = true;
						
		}
		//
		
	}


	print "done ($nOp Declarations, $nConst Constants)\n";
	fclose($fp);	
	
}//end: read_common_asm


$g_ITP = read_def('scriptdata/itp.def');
$g_MOBNAME = read_def('scriptdata/mobname.def');
$g_SKILL = read_def('scriptdata/skill.def');
read_enum_v2();
read_common_asm();

