<?php

require_once 'tables.php';
require_once 'npclist.php';
require_once 'globalvar.php';

$g_ERR = null;

define('WARN_MULTI_VAR_DECLARE', false);
define('WARN_CASE_TERM_BY_RETURN', false);	// Show Warning when choose case  gets terminated using return and lacks break

// 
// 
//
function scSplit($str){
	global $g_ERR;
	
	$ret = array();
	
	$len = strlen($str);
	$token = '';
	
	for($i = 0; $i < $len; $i++){
		$c = $str[$i];
		
		
		if($c === ' ' || $c === "\t"){
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}			
			continue;
		}else if($c === '/' && $i+1 < $len && $str[$i+1] === '/'){ // comment
			break; // we're done with this line.
			
		}else if($c === '(' || $c === '+' || $c === ')' || $c == '[' || $c == ']'){
			// 
			// [] () and concat + outsinde from string are also sepperated, to make parsing easier.
			//
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}
			
			$ret[] = $c;
		}else if($c === '/' || $c === '*' || $c === '+' || $c === '-'){
			//
			// Standart Arith. Operators
			//	
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}		
			
			$ret[] = $c;
		
		}else if( $c === '>' ){
			//
			// GR, GR_EQ
			// >, >= 
			//
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}
			
			if($i+1 < $len && $str[$i+1] === '='){
				$i++;
				$c .= '=';

			}
			
			$ret[] = $c;				

		}else if( $c === '=' ){
			//
			// LIKE EQ_GR EQ_LE
			// =, ==, =<, =>
			//
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}
			
			if( ($i+1) < $len ){
				$j = $str[$i+1];
				if( $j === '=' ){ // ==
					$i++;
					$c .= '=';
				}else if( $j === '<' ){ // =< 
					$i++;
					$c .= '<';
				}else if( $j === '>' ){ // =>
					$i++;
					$c .= '>';
				}

			}
			
			$ret[] = $c;				
		
		}else if( $c === '<' ){
			//
			// LE, LE_EQ
			// <, <=
			//
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}
			
			if($i+1 < $len && $str[$i+1] === '='){
				$i++;
				$c .= '=';

			}
			
			$ret[] = $c;				
		
		}else if( $c === '!' && ($i+1 < $len && $str[$i+1] === '=') ){
			//
			// NEQ (!=)
			//		
			if(strlen($token) > 0){
				$ret[] = $token;
				$token = '';
			}

			$i++; // double char!
			
			$ret[] = '!=';

		
		}else if($c === '"'){
			//
			// String! 
			//
			$offset_start = $i;
			$offset_max = $len-1;
			
			$token .= $c;
			
			while(1){

				//
				if($i == $offset_max){
					//
					// unterminated string! :(
					//
					$ret[] = $token;
					$token = '';
					
					$g_ERR = "Unterminated String Constant (String Begin at Offset $offset_start)";
					return false;
					
					break;
				}


				$c = $str[++$i];
				if($c === '"'){
					$token .= '"';
					$ret[] = $token;
					$token = '';
					
					break;
				}else{
					$token .= $c;
					
				}
					

			}//endwhile() - seek string termination "
			
		}else{
		
			$token .= $c;
		}	
	
	}//endfor() 


	if(strlen($token) > 0){
		$ret[] = $token;
		$token = '';
	}

	
return $ret;	
}//end scSplit


function frame($type, $line, $param){
	//
	// returns new 'stack frame' array for tokenizer.. (internal function)
	//
	
	return array('type' => $type, 
				 'line' => $line, 
				 'vars' => array(),
				 'param' => $param);
				 
}//end frame()


function clearVars(&$vars, $fe){
	
	// 
	foreach($fe['vars'] as $varName => &$varData){
		if(!array_key_exists($varName, $vars)){ // paranoid check. [internal error]
			echo "error.\n";
			fprintf(STDERR, "Internal Error, Variable '$varName' declared on frame (begin line {$fe['line']}) but unknown in vars array\n");
			die();
		}

		unset($vars[$varName]);
	}
	
}//end: clearVars()


function printFrameStack(SplStack &$stack, $showVars = false){

	fprintf(STDERR, "Frame Stack Overview\n");
	foreach($stack as $k => $v){
		fprintf(STDERR, " [$k] => %-8s {Opened in line: %u}\n", $v['type'], $v['line'] );
		if($showVars === true){
			foreach($v['vars'] as $varName => &$varData){
				fprintf(STDERR, "    Var %-12s {Declared in line: %u}\n", $varName, $varData['line']);
			}
		}
	}
	
}//end printFrameStack()



function getNumber(&$localVars, &$arr){
	global $g_ERR, $g_CONST, $g_SKILL, $g_ITP, $g_MOBNAME, $g_OP;
	$ret = null;

	$param = array_shift($arr);
		
	// 
	if($param === '['){
		//
		// enforced CONST lookup
		//
		if(sizeof($arr) < 2){
			$g_ERR = "[] -> Const Lookup, missing token - requires [CONST]\n";
			return false;		                
		}
		
		$const = array_shift($arr);
		$closeBacket = array_shift($arr);
		
		if(!array_key_exists($const, $g_CONST)){
			$g_ERR = "[] -> unknown const '$const'\n";
			return false;
		}
		
		$ret = array(	'type' => 'const_get',
						'param' => $const,
						'value' => $g_CONST[$const]
					);
		
		if($closeBacket !== ']'){
			$g_ERR = "[] -> Expected ']', got '$closeBacket'\n";
			return false;
		}
		
	
	}else if(!ctype_digit($param)){
			
		if(array_key_exists($param, $localVars)){
			//
			// stack var
			//
			$ret = array(	'type' => 'var_get_int',
							'param' => $param
						);

		}else if(array_key_exists($param, $g_CONST)){
			//
			// global const
			//
			$ret = array(	'type' => 'const_get',
							'param' => $param,
							'value' => $g_CONST[$param]
							);

		}else if(array_key_exists($param, $g_ITP) ) {
			//
			// item const
			//
			$ret = array(	'type' => 'itemconst_get',
							'param' => $param,
							'value' => $g_ITP[$param]
						);
						
		}else if(array_key_exists($param, $g_MOBNAME) ){
			//
			// monster const
			//
			$ret = array(	'type' => 'monsterconst_get',
							'param' => $param,
							'value' => $g_MOBNAME[$param]
						);
						
		}else if(array_key_exists($param, $g_SKILL) ){
			//
			// skill const
			// 
			$ret = array(	'type' => 'skillconst_get',
							'param' => $param,
							'vlaue' => $g_SKILL[$param]
						);
						
		}else if(array_key_exists($param, $g_OP) ){
			//
			// call proc 
			//
			$op = $g_OP[$param];
			
			$code = resolve($op, $localVars, $arr);
			if($code === false)
				return false;
			
			$ret = array( 	'type' => 'callfunc_ret_int',
							'param' => $param,
							'code' => $code
						);				
			
		}else{
			//
			// UKNOWN Token
			//
			$g_ERR = "unknown token '$param'";
			return false;
		}
	
		
	}else{
		//
		// INT Constant from Source (text)
		//
		$ret = array(	'type' => 'int',
						'param' => (int) $param
					);		
	}

return array($ret);
}//end: getNumber()


function getString(&$localVars, &$arr){
	global $g_ERR, $g_OP;
	$ret = array();
	
	// string has special case as it can be concated... 
	$expectString = true;
	while(1){
		if(sizeof($arr) == 0)
			break;
			
		$top = $arr[0];
		
		if($expectString === true && ($top[0] === '"' && $top[strlen($top)-1] === '"') ){
			$param = array_shift($arr);
			$ret[] = array( 'type' => 'str',
							'param' => $param
						);
						
			$expectString = false;
		
		}else if($expectString === true && ctype_digit($top) ){
			//
			// INT ? - invalid.
			//
			array_shift($arr);
			$g_ERR = "Expected String, got Integer";
			return false;
			
		}else if($expectString === true && ctype_alnum($top[0]) ){
			// constant ..
			$param = array_shift($arr);
			
			//
			// [FWI]: as far as i know, aegis only accepts string, function, localVars as param for 's' type param.
			//	
			if(array_key_exists($param, $g_OP)){
				//
				// call
				//
				$op = $g_OP[$param];
				$code = resolve($op, $localVars, $arr);
				if($code === false)
					return false;
				
				$ret[] = array( 'type' => 'callfunc_ret_str', 
								'param' => $param,
								'code' => $code
							);
			
			}else if(array_key_exists($param, $localVars)){
				//
				// stack var 9var_get0
				//
				$ret[] = array(	'type' => 'var_get_str',
								'param' => $param
							);
				
			}else{
				//
				// unknown
				//
				$g_ERR = "Unknown Token '$param'";
				return false;
			}
			
				
			$expectString  = false;
			
		}else if($top[0] === '+') {
			// concat!
			array_shift($arr);
			
			$ret[] = array( 'type' => 'concat_str',
							'param' => null 
						);
			
			$expectString  = true;
		}else{
			break;
		}	
	}//endwhile
	
	return $ret;
}//end: getString()


function getExpression(&$localVars, &$arr){
	global $g_ERR, $g_OP, $g_ITP, $g_MOBNAME, $g_SKILL, $g_CONST;

	// note:
	// this expression 'parser' 
	// wont re-order anything..
	// 

	//
	// Expression Note:
	//  expression means until end of line @ Aegis 
	// 	with other wirds, until arr is empty :) 
	//
	//
	// Expressions may have function calls
	// Expressions may also use stack vars and / or consts of any kind!
	// Expressions also have string constants as well as int consts 
	//
	// This Expression parser will count braces. 
	//

	$expr = array();
	$curexpr = array();

	$braceLevel = 0;
	$i = 0;
	while( sizeof($arr) != 0 ){
		$param = array_shift($arr);

		if($param === '('){
			$braceLevel++;
			
		}else if($param === ')') {
			if(--$braceLevel < 0){
				$np = $i+1;
				$g_ERR = "Unexpected ')' (Token Offset $np)";
				return false;
			}
			
		}else if(ctype_digit($param)){
			//
			// Integer Constant
			// 
			$curexpr[] = array('type' => 'int',
								'param' => $param
							);
							
		}else if($param[0] === '"' && $param[strlen($param)-1] === '"'){
			//
			// String Constant
			//
			// 
			$curexpr[] = array('type' => 'str',
								'param' => $param
							);			
			
		}else if($param === '['){
			//
			// const lookup
			// npcv "asd#1" [VAR_MYMOBCOUNT]
			//
			// next token is CONST and after that its ]
			// only vars in g_CONST are supported.
			//
			if(sizeof($arr) < 2){
				$np = $i+1;
				$g_ERR = "[] -> Const Lookup, missing token - requires [CONST] (Token Offset $np)\n";
				return false;
			}

			$const = array_shift($arr);
			$closeBackett = array_shift($arr);
			
			if(!array_key_exists($const, $g_CONST)){
				$np = $i+1;
				$g_ERR = "[] -> Const Lookup, unknown constant '$const' (Token Offset $np)\n";
				return false;
			}
			
			
			if($closeBackett !== ']'){
				$np = $i+1;
				$g_ERR = "[] -> expected ], got '$closeBackett' (Token Offset $np)\n";
				return false;	
			}
			
			$curexpr[] = array('type' => 'const_get',
								'param' => $const,
								'value' => $g_CONST[$const]
							);
			
			
		}else if($param === '==' || $param === '!=' || $param === '>=' || $param === '<=' || $param === '=<' || $param === '=>' || $param === '>' || $param === '<' || $param === '&' || $param === '|' || $param === '+' || $param === '-' || $param === '/' || $param === '*'){
			//
			// Operator
			//
			
			// AND / OR (logical) aegis has _no_ binary operators. 
/*			if($param === '&' || $param === '|'){
				$expr[] = $curexpr;
				$curexpr = array();
				
				$expr[] = $param;

			}else{
			*/
				//
				//
				//
				$curexpr[] = array('type' => 'operator',
									'param' => $param
								);
//			}
					
		}else{
			// 
			// Constant, try lookup :)
			//
			//print "const! $param\n";
			
			if(array_key_exists($param, $g_OP)){
				//
				// call
				//				
				$op = $g_OP[$param];

				$code = resolve($op, $localVars, $arr);
				
				$curexpr[] = array( 'type' => 'callfunc_ret', // need 2nd pass for reorder / type of operation (depends on operator)
									'param' => $param,
									'code' => $code
								);
				
			}else if(array_key_exists($param, $localVars)){
				//
				// stackvar
				//
				$curexpr[] = array( 'type' => 'var_get', 
									'param' => $param,
								);
				
			}else if(array_key_exists($param, $g_CONST)){
				//
				// Global CONST 
				//
				$curexpr[] = array( 'type' => 'const_get',
									'param' => $param,
									'value' => $g_CONST[$param]
								);
				
			}else if(array_key_exists($param, $g_ITP)){
				//
				// ITP const
				//
				$curexpr[] = array( 'type' => 'itemconst_get', 
									'param' => $param,
									'value' => $g_ITP[$param]
								);
				
			}else if(array_key_exists($param, $g_MOBNAME)){
				//
				// MOBNAME const 
				//
				$curexpr[] = array(	'type' => 'monsterconst_get',
									'param' => $param,
									'value' => $g_MOBNAME[$param]
								);
			
			}else if(array_key_exists($param, $g_SKILL)){
				//
				// SKILL const
				//
				$curexpr[] = array( 'type' => 'skillconst_get',
									'param' => $param,
									'value' => $g_SKILL[$param]
								); 
				
			}else{
				//
				// -- UNKNWON --
				//
				$np = $i+1;
				$g_ERR = "Unexpected '$param' (Token Offset $np)";
				return false;
			}
			
		}
		
				
		$i++;
	}//endwhile 
	
	// 
	// move last expr to expr 
	// 
	$expr[] = $curexpr;
	

	return $expr;
}//end: getExpression()



/* Resolves Code to Tokenlist (code) */
function resolve($op, &$localVars, &$arr){
	global $g_ERR;
	
	// param number ..
	$szArr = sizeof($arr);
	if($szArr < $op['nParam']){
		if($op['va']===true) $atleast = ' at least';
		else $atleast = '';
		
		$g_ERR = "{$op['name']} expects$atleast {$op['nParam']} Parameters, $szArr given\n";
		
		return false;
	}

	//
	$funcParams = array();
	
	//
	// Support [] one param funcs. 
	//
	if($op['nParam'] === 1 && $op['va'] === false && $szArr >= 3 && ($arr[0] === '[' && $arr[2] === ']')  ) {
		array_shift($arr); // remove [
		$expectCloseBacket = true;
	}else
		$expectCloseBacket = false;
		
	
	for($i = 0; $i < $op['nParam']; $i++){
		$type = $op['param'][$i];
		$tmp = null;
		switch($type){
			case 'n': // num
				$tmp = getNumber($localVars, $arr);
			break;
			
			case 's': // str
				$tmp = getString($localVars, $arr);
			break;
			
			case 'e': // expr (if/while etc)
				$tmp = getExpression($localVars, $arr);
			break;
			
			default:
				$ni = $i+1;
				$g_ERR = "Internal Error, '{$op['name']} has unknown type definition [$type] for parameter $np\n";
				return false;
			break;
		}
		
		if($tmp === false){
			$np = $i+1;
			$g_ERR = "Parser Error, '{$op['name']}', parameter $np (type $type)\n Parser Message: $g_ERR\n";
			return false;
		}
		
		$funcParams[] = $tmp;					
	}
	
	// va?
	if($op['va'] === true){
		
		while(sizeof($arr) != 0){
			$tmp = null;
			switch($type){
				case 'n': //num
					$tmp = getNumber($localVars, $arr);
				break;
				
				case 's': //str
					$tmp = getString($localVars, $arr);
				break;
				
				default:
					$np = $i+1;
					$g_ERR = "Parser Error, '{$op['name']} has unknown type definition [$type] for VA parameter $np\n";
					return false;
				break;
			}//endswitch
			
			if($tmp === false){
				$np = $i+1;
				$g_ERR = "Parser Error, '{$op['name']}', VA parameter $np (type $type)\n Parser Message: $g_ERR\n";
				return false;
			}
			
			$funcParams[] = $tmp;
			
			$i++;
		}//endwhile
	}//endif: va?
	
	
	//
	// support for one param ops with [] syntax:
	//
	if($expectCloseBacket === true){
		if(sizeof($arr) <= 0 || $arr[0] !== ']'){
			$g_ERR = "Syntax Error, expected ]";
			return false;
		}
		
		array_shift($arr);		
	}
	
//	return array(	'type' => 'callfunc',
//					'param' => $op['name'],
//					'code' => $funcParams
//				);

	return $funcParams;	
}//end: resovle



///
function tokenize($fileName){
	global $g_ERR, $g_OP;
	
	print "Parsing $fileName... ";
	
	$fp = @fopen($fileName, 'r');
	if(!$fp)
		die("failed (IO open Error)\n");

	//
	$tl = array();
	
	
	//
	$frames = new SplStack();
	$frames->setIteratorMode(SplDoublyLinkedList::IT_MODE_LIFO | SplDoublyLinkedList::IT_MODE_KEEP);
	
	$vars = array();
	
	$op = null;
	$last_op = null;
	$last_cmd = null;
	$linecnt = 0;
	while(1){
		$linecnt++;
		$line = fgets($fp, 8192);
		if($line === false)
			break;
		
		$line = trim($line);
		$linelen = strlen($line);
		
		if($linelen === 0)
			continue;
		
		// Aegis Supports comments as C++ style and somehow ini style ;
		$c = $line[0];
		if($c === ';' || ($c === '/' && $line[1] === '/'))
			continue;
		
		
		// Support for C style comment
		$c89comment_isString = false;
		for($i = 0; $i < $linelen-1; $i++){
		
			if($line[$i] === '"'){
				if($c89comment_isString  === false)
					$c89comment_isString = true;
				else
					$c89comment_isString = false;
			}else
			if($c89comment_isString == false && ($line[$i] === '/' && $line[$i+1] === '*') ){
				$offset_start = $i;
				$offset_end = 0;
				$i+=2;

				// seek for */ 
				$terminated = false;
				for( ; $i < $linelen-1; $i++){
					if($line[$i] === '*' && $line[$i+1] === '/'){
						$terminated = true;
						$offset_end = $i+2;
						break;
					}
				}
				
				
				if($terminated === true){
					//
					// terminated in 'place' (line)
					// 
					// cut out the comment and contiue parsing!
					//
					$line = substr($line, 0, $offset_start).substr($line, $offset_end);
					$linelen = strlen($line);

					if($linelen === 0)
						goto _outercontinue; // nothing to parse left.
					
					break; // break outer for!					
				}else{
					//
					// unterminated, fetch linesu ntil termination!
					//
					$start_line = $linecnt;
					$offset_end = 0;
					while(1){
						$linecnt++;
						$line = fgets($fp, 8192);
						if($line === false)
							break;
						
						$line = trim($line);
						$linelen = strlen($line);
						
						if($linelen === 0)
							continue; // fetch next!
						
						// as we're in comment, we dont have to check for C++ (//) style comment! 
						
						for($i = 0; $i < $linelen-1; $i++){
							// NO Check for in string required, as we're in comment!
							if($line[$i] === '*' && $line[$i+1] === '/'){
								$terminated = true;
								$offset_end = $i+2;
								goto _c89comment_endseek_done;
							}
						}
						
						
					}
					_c89comment_endseek_done:
					
					// :(
					if($terminated === false){
						print "error.\n";
						fprintf(STDERR, "Error: unterminated C-Style comment (Started in line $start_line)\n");
						die();
						
					}else{
						//
						// cut line!
						//
						if($offset_end >= $linelen)
							goto _outercontinue; // nothgin to parse left
						
						$line = substr($line, $offset_end);
						$linelen = strlen($line);
						
						if($linelen == 0)
							goto _outercontinue; // nothing to parse left.
					}
					
					break; // break outer for!	
				}
				
			}
		}
		//end support for c-style comment
		
		
		// Split Line in simple Tokens
		$arr = scSplit($line);	
 		if($arr === false){
			print "error.\n";
			
			fprintf(STDERR, "$fileName:$linecnt: $g_ERR\n");
			die();
		}
		

		// Op To handle
		$op = null;
		
		
		//
		$cmd = array_shift($arr);
		if(strcmp($cmd, 'if') === 0){
			//
			// if
			//
			$frames->push( frame('if', $linecnt, $arr) );
			
			$op = array('name' => 'if',
						'blockcheck' => false,
						'param' => 'e',
						'nParam' => 1,
						'va' => false,
						'va_type' => null);

		}else if(strcmp($cmd, 'elseif') === 0){
			//
			// elseif
			//
			if( $frames->count()  === 0 ){
				print "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: unexpected elseif (no open if/elseif frame)\n");
				die();
			}
			
			$fe = $frames->pop();
			$type = &$fe['type'];
			
			if($type !== 'if' && $type !== 'elseif'){
				print "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unxpected elseif (no open if/elseif frame) [Current Frame: $type]\n\n");
				$frames->push($fe);
				printFrameStack($frames);
				die();
			}
			
			clearVars($vars, $fe);
			
			
			// ok		
			$frames->push( frame( 'elseif', $linecnt, $arr) );

			$op = array('name' => 'elseif',
						'blockcheck' => false,
						'param' => 'e',
						'nParam' => 1,
						'va' => false,
						'va_type' => null);
				
				
		}else if(strcmp($cmd, 'else') === 0){
			//
			// else
			//
			if( $frames->count() === 0 ){
				echo "error.\n"; 
				fprintf(STDERR, "$fileName:$linecnt: unexpected else (no open if/elseif frame)\n");
				die();
			}
			
			$fe = $frames->pop();
			$type = &$fe['type'];
			if($type !== 'if' && $type !== 'elseif'){
				print "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected else (no open if/elseif frame) [Current Frame: $type]\n\n");
				$frames->push($fe);
				printFrameStack($frames);
				die();
			}
			
			clearVars($vars, $fe);
			
			// ok
			$frames->push( frame( 'else', $linecnt, $arr) );

			$op = array('name' => 'else',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);

		
		}else if(strcmp($cmd, 'endif') === 0){
			//
			// endif
			//
			if( $frames->count() === 0 ){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected endif (no open if/elseif/else frame)\n");
				die();
			}
			
			$fe = $frames->pop();
			$type = &$fe['type'];
			if($type !== 'if' && $type !== 'elseif' && $type !== 'else'){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected endif (no open if/elseif/else frame) [Current Frame: $type]\n\n");
				$frames->push($fe);
				printFrameStack($frames);
				die();
			}
			
			clearVars($vars, $fe);
			
			$op = array('name' => 'endif',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);
			
		}else if(strcmp($cmd, 'while') === 0){
			//
			// while
			//
			
			$frames->push( frame('while', $linecnt, $arr) );

			$op = array('name' => 'while',
						'blockcheck' => false,
						'param' => 'e',
						'nParam' => 1,
						'va' => false,
						'va_type' => null);

			
		}else if(strcmp($cmd, 'exitwhile') === 0){
			//
			// exitwhile
			//
			if( $frames->count() === 0 ){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected exitwhile (no open while frame)\n");
				die();
			}
			
			// Search open while frame (exitwhile without while is invalid(
			$fe = null;
			foreach( $frames as $k => $v ){
				if($v['type'] === 'while'){
					$fe = $v;
					break;
				}
			}
			
			if($fe === null){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected exitwhile (no open while frame)\n");
				printFrameStack($frames);
				die();
			}
	

			$op = array('name' => 'exitwhile',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);

			
		}else if(strcmp($cmd, 'endwhile') === 0){
			//
			// endwhile
			//
			if( $frames->count() === 0 ){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected endwhile (no open while frame)\n");
				die();
			}
			
			$fe = $frames->pop();
			$type = &$fe['type'];
			
			if($type !== 'while'){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected endwhile (no open while frame) [Current Frame: $type]\n\n");
				$frames->push($fe);
				printFrameStack($frames);
				die();
			}
			
			clearVars($vars, $fe);

			$op = array('name' => 'endwhile',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);

		
		}else if(strcmp($cmd, 'choose') === 0){
			//
			// choose
			//
			
			$frames->push( frame('choose', $linecnt, $arr) );

			$op = array('name' => 'choose',
						'blockcheck' => false,
						'param' => 'n',
						'nParam' => 1,
						'va' => false,
						'va_type' => null);

		
		}else if(strcmp($cmd, 'case') === 0){
			//
			// case
			//
			if( $frames->count() === 0 ){
				echo "error.\n";
				
				fprintf(STDERR, "$fileName:$linecnt: unexpected case (no open choose frame)\n");
				die();
			}
			
			_case_fetchframe:
			$fe = $frames->top();
			$type = &$fe['type'];
			
			
			if($type === 'case'){	// HACK: to support closing the case by using return
				if($last_cmd === 'return'){
					if(WARN_CASE_TERM_BY_RETURN)
						fprintf(STDERR, "Warning: $fileName:$linecnt: terminating CASE by using return! (Recommendation: add break after return)\n");
				
					$fe = $frames->pop();
					clearVars($vars, $fe);
					goto _case_fetchframe;
					
				}else
					goto _case_errtype;

			}else if($type !== 'choose'){
				_case_errtype:
				echo "error.\n";
				
				if($type === 'case'){
					// aegis script has no support for 'fallthru' multiple cases, 
					// each case MUST be terminated,
					// otherwise this could end in unexpected behaviour.
					fprintf(STDERR, "$fileName:$linecnt: expected break, CASE fall-thru is not supported in the Aegis scripting Language!\n");
					fprintf(STDERR, "\nNote:\n Each case must be terminated using break,\n otherwise the script may have a unexpected behavior!\n\n");
					printFrameStack($frames);
					die();
					
				}else{
					// 
					fprintf(STDERR, "$fileName:$linecnt: unexpected case (no open choose frame) [Current Frame: $type]\n\n");
					$frames->push($fe);
					printFrameStack($frames);
					die();
				}
			}//
			
			// ok
			$frames->push( frame('case', $linecnt, $arr) );


			$op = array('case' => 'case',
						'blockcheck' => false,
						'param' => 'n',
						'nParam' => 1,
						'va' => false,
						'va_type' => null);

			
		}else if(strcmp($cmd, 'break') === 0){
			// 
			// break
			//
			if( $frames->count() === 0 ){
				echo "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: unexpectd break (no open case frame)\n");
				die();
			}
			
			$fe = $frames->pop();
			$type = &$fe['type'];
			if($type !== 'case'){
				echo "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: unexpected break (no open case frame) [Current Frame: $type]\n");
				die();
			}
			
			clearVars($vars, $fe);


			$op = array('name' => 'break',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);


		}else if(strcmp($cmd, 'endchoose') === 0){
			//
			// endchoose
			//
			if( $frames->count() === 0){
				echo "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: unexpected endchoose (no open choose frame)\n");
				die();
			}
			
			_endchoose_fetchframe:
			$fe = $frames->pop();
			$type = &$fe['type'];
			
			if($type === 'case'){	// HACK: to support closing the case by using return
				if($last_cmd === 'return'){
					if(WARN_CASE_TERM_BY_RETURN)
						fprintf(STDERR, "Warning: $fileName:$linecnt: terminating CASE by using return! (Recommendation: add break after return)\n");

					clearVars($vars, $fe);
					goto _endchoose_fetchframe;
				}else
					goto _endchoose_errtype;
				
			}else if($type !== 'choose'){
				_endchoose_errtype:
				echo "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: unexpected endchoose (no open choose frame) [Current Frame: $type]\n\n");
				
				$frames->push( $fe );
				printFrameStack($frames);
				die();
			}
			
			clearVars($vars, $fe);

			$op = array('name' => 'endchoose',
						'blockcheck' => false,
						'param' => '.',
						'nParam' => 0,
						'va' => false,
						'va_type' => null);

			
		}else if(strcmp($cmd, 'var') === 0){
			//
			// var
			//
			// var x [= expr]
			if(sizeof($arr) < 1){
				echo "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: var expects at least 1 Parameter (varname)\n");
				die();
			}
			
			$frameNum = $frames->count();
			
			$varName = array_shift($arr);
			$varData = array('frame' => $frameNum, 'line' => $linecnt);
			
			// add to lookup array.
			if(array_key_exists($varName, $vars)){
				if(WARN_MULTI_VAR_DECLARE)
					fprintf(STDERR, "Warning: $fileName:$linecnt: var '$varName' already defined (line: {$vars[$varName]['line']})\n");
	
			}else{
				//
				$vars[ $varName ] = $varData;
			
			
				if($frameNum > 0){
					$fe = $frames->pop();
					$fe['vars'][$varName] = $varData;
					$frames->push($fe); // this is typical for php....
				}
			}		
			
			//
			if(sizeof($arr) == 1){
				// syntax error 
				echo "error.\n";
				
				fprintf(STDERR, "$fileName;$linecnt: Syntax Error in var expression\n");
				fprintf(STDERR, " Line: $line\n");
				
			}else if(sizeof($arr) > 1){
			
				$param = 'e';
				$nParam = 1;
			}else{
				$param = '.';
				$nParam = 0;
			}
			

			$op = null;
			/*
			$op = array('blockcheck' => false,
						'param' => $param,
						'nParam' => $nParam,
						'va' => false,
						'va_type' => null);

			*/
		}else{
			//
			// Other Token (cmd or var expression)
			//
			
			if(array_key_exists($cmd, $vars) ){
				// skip line as thers no expr. parser right now ..
				continue;
							
			}else{


				if(!array_key_exists($cmd, $g_OP)){
					print "error.\n";
					fprintf(STDERR, "$fileName:$linecnt: unknown cmd '$cmd'\n");
					fprintf(STDERR, " Line: $line\n");
					die();            
				}else{
				
					$op = $g_OP[$cmd];
				}
			}					
				
		}

			
		if($op !== null){
			
			
			//
			// Block Check
			// 
			if($op['blockcheck'] === true){
				// Frame Stack MUST be empty.
				if($frames->count() !== 0){
					print "error.\n";
					fprintf(STDERR, "$fileName:$linecnt: blockcheck failed ($cmd expects the frame-stack to be empty)\n\n");
					fprintf(STDERR, "Note:\n This may be re result of a missing endif/endwhile/endchoose\n\n");

					printFrameStack($frames);
					die();
				}
			
				/// otherwise each blockcheck clears var!
				unset($vars);
				$vars = array();

			}//endif: block check
			
			//
			// Check Op parameters
			//
			$szArr = sizeof($arr);
			if($szArr < $op['nParam']){
				print "error.\n";

				if($op['va'] === true) $atleast = ' at least';
				else $atleast = '';

				fprintf(STDERR, "$fileName:$linecnt: Syntax Error, '$cmd' expects$atleast {$op['nParam']} Parameter, $szArr given\n");
				fprintf(STDERR, " Line: $line\n");
				die();
			}
			
			$code = resolve($op, $vars, $arr);
			if($code === false){
				print "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: $g_ERR\n");
				fprintf(STDERR, " Line: $line\n");
				die();
			}
			
			if(sizeof($arr) !== 0){
				print "error.\n";
				fprintf(STDERR, "$fileName:$linecnt: Syntax Error, unexpected Argument left\n");
				fprintf(STDERR, " Line: $line\n");
				die();
			}
			
			//
			$inst = array(	'type' => 'callfunc',
							'param' => $cmd,
							'code' => $code
						);
			
			//
			// add to token List 
			//
			$tl[] =  array( 'line' => $linecnt,
								'code' => $inst
							);
					
			
		}
		
		
		$last_cmd = $cmd;
		$last_op = $op;	
		_outercontinue:
		unset($arr, $line);
	}//endwhile()

	//
	// Lets do a BlockCheck..
	//
	if($frames->count() !== 0){
		print "error.\n";
		fprintf(STDERR, "$fileName:$linecnt: blockcheck failed (unterminated frames)\n\n");
		fprintf(STDERR, "Note:\n This may be re result of a missing endif/endwhile/endchoose\n\n");
		printFrameStack($frames);
		die();
	}
	
	print "done.\n";
	fclose($fp);
	
	return $tl;
}//end: tokenize()


//
// Example: 
//  just parse the npcs 
//  (LINT)
//


$thisctime = filectime($argv[0]);
foreach($g_LIST as $i => &$sc){

	$cachefile = './cache/'.hash('sha512', "$thisctime ".$sc).'.tcache';
	$ctime_sc = filectime($sc);
	$ctime_cache = 0;
		
	if(file_exists($cachefile))
		$ctime_cache = filectime($cachefile);

	//
	if($ctime_cache <= $ctime_sc){
		$data = tokenize($sc);
		file_put_contents($cachefile, serialize($data) );
	}else{
		//$data = unserialize( file_get_contents($cachefile) );
	}
	
	unset($data);

}



//$tl = tokenize('npcdata/person/matnani/smile_event.sc');
//print_r($tl);

//$tl = tokenize("test.sc");
//print_r($tl);
